#!/usr/bin/python

'''
This script computes the monthly climatology from IMERG and saves them in
netCDF4 format and, if flagged, GeoTIFF format. It uses the daily IMERG 
netCDF4 files (GPM_3IMERGD*) as input, organized under yearly directories 
under the path specified by "imergdaypath" below.

This script requires Python 3.6 or later. On top of modules from the 
Python Standard Library, this script also requires NumPy and netCDF4, 
and, for GeoTIFF format, rasterio. This script has been tested on 
Python 3.11.4, NumPy 1.25.2, netCDF4 1.6.2, and rasterio 1.2.10.

Jackson Tan (jackson.tan@nasa.gov)
2024 03 29
'''

def main():

    import numpy as np
    import netCDF4 as nc4
    import sys
    import os
    from calendar import monthrange, month_name
    from datetime import datetime

    year0, year1, run = int(sys.argv[1]), int(sys.argv[2]), sys.argv[3]
    ver = 'V07B'    # version number in the GES DISC daily filenames
    make_geotiff = True

    if make_geotiff:
        import rasterio

    runname = {'F': 'Final', 'L': 'Late', 'E': 'Early'}

    # CHANGE THE PATHS TO POINT TO THE FILES
    imergdaypath = '/path/to/daily/IMERG/netCDF4/files/'
    outpath = '/path/to/output/'
    author = 'Name (email)'

    os.makedirs(outpath, exist_ok = True)

    # Define the grid information.

    nlon, nlat = 3600, 1800
    lonedges = np.linspace(-180, 180, nlon + 1, dtype = 'f8')
    latedges = np.linspace(-90, 90, nlat + 1, dtype = 'f8')
    lons = lonedges[1:] - 0.5 * np.diff(lonedges)
    lats = latedges[1:] - 0.5 * np.diff(latedges)

    # Define miscellaneous settings.

    threshFrac = 0.9    # mask data if available input is below this fraction

    fv = -9999.9
    timestamp = 'S000000-E235959'
    runname = {'F': 'Final', 'L': 'Late', 'E': 'Early'}

    # Read the data and package it into a netCDF4 file.

    for month in range(1, 13):

        infiles = []
        psum = np.zeros([nlon, nlat], dtype = 'f8')
        pcnt = np.zeros([nlon, nlat], dtype = 'i8')

        # read the daily files across the years for the month
        nday = 0
        for year in range(year0, year1 + 1):
            for day in range(1, monthrange(year, month)[1] + 1):
                if   run == 'L' or run == 'E':
                    file = (f'3B-DAY-{run}.MS.MRG.3IMERG.{year:4d}{month:02d}'
                            f'{day:02d}-{timestamp}.{ver}.nc4')
                elif run == 'F':
                    file = (f'3B-DAY.MS.MRG.3IMERG.{year:4d}{month:02d}'
                            f'{day:02d}-{timestamp}.{ver}.nc4')
                infiles.append(file)
                with nc4.Dataset(f'{imergdaypath}{year}/{file}', 'r') as f:
                    p = np.ma.filled(f['precipitation'][0, :], fv)
                pcnt += (p != fv)
                psum[p != fv] += p[p != fv]
                nday += 1

        # average the precipitation, retaining units as mm / day
        P = np.divide(psum, pcnt, where = (pcnt >= nday * threshFrac),
                      out = np.full([nlon, nlat], fv, 'f4'))
        P = np.ma.masked_values(P, fv)

        # open the output netCDF4 file
        filename = (f'{outpath}IMERG-{runname[run]}.CLIM.'
                    f'{year0}-{year1}.{month:02d}.{ver}.nc4')
        f = nc4.Dataset(filename, 'w', format = 'NETCDF4')
        v = {}

        # set the dimensions
        lat = f.createDimension('lat', len(lats))
        lon = f.createDimension('lon', len(lons))

        # write the latitude data and attributes
        v['lat'] = f.createVariable('lat', 'f4', ('lat',), zlib = True)
        v['lat'][:] = lats
        v['lat'].units = "degrees_north"
        v['lat'].long_name = 'Latitude'

        # write the longitude data and attributes
        v['lon'] = f.createVariable('lon', 'f4', ('lon',), zlib = True)
        v['lon'][:] = lons
        v['lon'].units = "degrees_east"
        v['lon'].long_name = 'Longitude'

        # write the precipitation data and attributes
        v['precipitation'] = f.createVariable('precipitation', 'f4', 
                                              ('lon', 'lat'), 
                                              fill_value = fv,
                                              zlib = True)
        v['precipitation'][:] = P
        v['precipitation'].long_name = 'climatological precipitation'
        v['precipitation'].units = 'mm / day'
        v['precipitation'].coordinates = 'lat lon'

        # write the global attributes
        f.title = (f'IMERG {runname[run]} Monthly Climatological '
                   f'Precipitation ({year0} to {year1})')
        f.author = author
        f.description = (f'Average of all valid daily precipitation in ' 
                         f'{month_name[month]} ' 
                         f'from {year0} to {year1} at every 0.1 deg.')
        f.version = ver
#        f.history = (f'File created on '
#                     f'{datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")} '
#                     f'UTC. Inputs: {";".join(infiles)}')
        f.history = (f'File created on '
                     f'{datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")} UTC.')
        f.source = (f'IMERG {runname[run]} Daily Precipitation '
                    f'(10.5067/GPM/IMERGD{run}/DAY/07)')
        f.comment = (f'A value is given only in a grid box if at least ' 
                     f'{int(threshFrac * 100):d}% ' 
                     f'of the data over the record is available. ' 
                     f'This product is created as a convenience to users. ' 
                     f'Please verify the data and use them at your own ' 
                     f'risk. Direct any queries to '
                     f'https://gpm.nasa.gov/contact.')
        f.Conventions = 'CF-1.8'
        f.close()

        if make_geotiff:

            # calculate the GeoTIFF transformation.
            ll = (lons[ 0], lats[ 0])
            ul = (lons[ 0], lats[-1])
            lr = (lons[-1], lats[ 0])
            ur = (lons[-1], lats[-1])
            transform = rasterio.transform.from_gcps([
                    rasterio.control.GroundControlPoint(0, 0, *ll), 
                    rasterio.control.GroundControlPoint(0, nlat, *ul), 
                    rasterio.control.GroundControlPoint(nlon, 0, *lr), 
                    rasterio.control.GroundControlPoint(nlon, nlat, *ur)
                ])

            # write to GeoTIFF file
            filename = (f'{outpath}IMERG-{runname[run]}.CLIM.'
                        f'{year0}-{year1}.{month:02d}.{ver}.tif')
            with rasterio.open(filename, 'w', driver = 'GTiff',
                               height = P.shape[1], width = P.shape[0],
                               count = 1, dtype = P.dtype,
                               crs = '+proj=latlong',
                               transform = transform,
                               compress = 'lzw') as f:
                f.write(P.T[::-1], 1)

if __name__ == "__main__":
    main()